"""Data types related to doing queries."""
from iobeam.utils import utils
from iobeam.resources import data

# alias query.TimeUnit to data.TimeUnit
TimeUnit = data.TimeUnit

class Query(object):
    """Represents a data query."""

    @staticmethod
    def _validIntOrRaise(val, msg):
        """Check if value is an int type

        Params:
            val - Value to be checked
            msg - ValueError msg if it is not an int

        Returns:
            True if `val` is int, False if is None.

        Raises:
            ValueError - If val is a non-int.
        """
        if val is None:
            return False
        elif not isinstance(val, int):
            raise ValueError(msg)
        else:
            return True

    @staticmethod
    def _validPositiveIntOrRaise(val, msg):
        """Check if value is an int and > 0.

        Params:
            val - Value to be checked
            msg - ValueError msg if it is not an int

        Returns:
            True if `val` is int and > 0, False if it is None.

        Raises:
            ValueError - If val is a non-int
        """
        if not Query._validIntOrRaise(val, msg):
            return False
        elif val > 0:
            return True
        else:
            raise ValueError(msg)

    def __init__(self, projectId, deviceId=None, seriesName=None,
                 timeUnit=TimeUnit.MILLISECONDS):
        """Construct a new query object."""
        utils.checkValidProjectId(projectId)
        if timeUnit is None:
            timeUnit = TimeUnit.MILLISECONDS
        if not isinstance(timeUnit, TimeUnit):
            raise ValueError("timeUnit must be a query.TimeUnit")

        self._pid = projectId
        self._did = deviceId
        self._series = seriesName
        self._params = {}
        self._timeUnit = timeUnit
        if timeUnit != TimeUnit.MILLISECONDS:
            self._params["timefmt"] = str(timeUnit.value)

    def getUrl(self):
        """Gets the resource locator portion of the exports API call

        Returns:
            String of format <projectId>/<deviceId or 'all'>/<seriesName or 'all'>
        """
        did = self._did or "all"
        series = self._series or "all"
        return "{}/{}/{}".format(self._pid, did, series)

    def getParams(self):
        """Return parameter dictionary for this query."""
        return self._params

    def limit(self, limit):
        """Sets limit of this query, i.e., the max # of results per series."""
        msg = "limit must be a positive int"
        if Query._validPositiveIntOrRaise(limit, msg):
            self._params["limit"] = limit
        return self

    def fromTime(self, time):
        """Sets the from time limit for results."""
        msg = "time must be an int ({} from epoch)".format(self._timeUnit.value)
        if Query._validIntOrRaise(time, msg):
            ts = data.Timestamp(time, unit=self._timeUnit)
            self._params["from"] = ts.asMilliseconds()
        return self

    def toTime(self, time):
        """Sets the to time limit for results."""
        msg = "time must be an int ({} from epoch)".format(self._timeUnit.value)
        if Query._validIntOrRaise(time, msg):
            ts = data.Timestamp(time, unit=self._timeUnit)
            self._params["to"] = ts.asMilliseconds()
        return self

    def inTimeRange(self, start, end):
        """Sets the time limits for results.

        Params:
            start - Beginning of time range to get results from
            end - End of time range to get results from
        """
        if start is not None and end is not None:
            self.fromTime(start).toTime(end)
            if end < start:
                self._params.pop("from", None)
                self._params.pop("to", None)
                raise ValueError("end cannot be less than start")
        return self

    def lessThan(self, value):
        """Sets the max value for values in the results."""
        msg = "value must be an int"
        if Query._validIntOrRaise(value, msg):
            self._params["less_than"] = value
        return self

    def greaterThan(self, value):
        """Sets the min value for values in the results."""
        msg = "value must be an int"
        if Query._validIntOrRaise(value, msg):
            self._params["greater_than"] = value
        return self

    # pylint: disable=redefined-builtin
    def inValueRange(self, min, max):
        """Sets the value range for results.

        Params:
            min - Minimum value to have in results
            max - Maximum value to have in results
        """
        if min is not None and max is not None:
            self.lessThan(max).greaterThan(min)
            if max < min:
                self._params.pop("less_than", None)
                self._params.pop("greater_than", None)
                raise ValueError("max value cannot be larger than min value")
        return self
    # pylint: enable=redefined-builtin

    def equals(self, value):
        """Sets the value that all results must equal."""
        msg = "value must be an int"
        if Query._validIntOrRaise(value, msg):
            self._params["equals"] = value
        return self



"""Used to communicate with iobeam's Imports API"""
from iobeam.endpoints import service
from iobeam.http import request
from iobeam.utils import utils


class ImportService(service.EndpointService):
    """Communicates with the backend and exposes available Imports API methods."""

    _BATCH_SIZE = 1000  # max # of pts in a single request

    def __init__(self, token, requester=None):
        service.EndpointService.__init__(self, token, requester=requester)

    @staticmethod
    def _makeRequest(projectId, deviceId, dataset):
        """Creates the body of a import request.

        Params:
            projectId - Project ID of the request
            deviceId - Device ID of the request
            dataset - The data series of the request, a map with names to lists
                of data.DataPoints.

        Returns:
            A dictionary that is the body of an import request.
        """
        sources = []
        req = {
            "project_id": projectId,
            "device_id": deviceId,
            "sources": sources,
            "timefmt": "usec"
        }
        for series in dataset:
            pts = []
            obj = {"name": series, "data": pts}
            for d in dataset[series]:
                pts.append(d.toDict())
            sources.append(obj)

        return req

    @staticmethod
    def _makeBatchRequest(projectId, deviceId, dataStore):
        """Creates the body of a table import request.

        Params:
            projectId - Project ID of the request
            deviceId - Device ID of the request
            dataStore - The DataStore that makes the body of the request.

        Returns:
            A dictionary that is the body of an import request.
        """
        sources = {}
        req = {
            "project_id": projectId,
            "device_id": deviceId,
            "sources": sources,
            "timefmt": "usec"
        }

        sources["fields"] = ["time"]
        for f in dataStore.columns():
            sources["fields"].append(f)
        sources["data"] = []
        for r in dataStore.rows():
            row = []
            for f in sources["fields"]:
                row.append(r[f])
            sources["data"].append(row)

        return req

    @staticmethod
    def _makeListOfReqs(projectId, deviceId, dataset):
        """Creates a list of import requests from a data set.

        If the data set is under _BATCH_SIZE, it will be one request. Otherwise
        it will be split into multiple requests as follows:
        (1) if a single series has less than ImportService._BATCH_SIZE points,
            it will be a request.
        (2) if a single series has more, it will be broken into multiple requests of
            ImportService._BATCH_SIZE size.

        Params:
            projectId - Project ID of the requests
            deviceId - Device ID of the requests
            dataset - The data set that will be broken into requests.

        Returns:
            A list of import request bodies.
        """
        totalLen = sum(len(dataset[k]) for k in dataset)

        reqs = []
        # No series, no requests
        if totalLen == 0:
            pass
        # Everything can fit in one request
        elif totalLen <= ImportService._BATCH_SIZE:
            reqs.append(
                ImportService._makeRequest(projectId, deviceId, dataset))
        # Need to create multiple requests
        else:
            for series in dataset:
                seriesLen = len(dataset[series])
                # If the series itself is small enough, send it
                if seriesLen <= ImportService._BATCH_SIZE:
                    temp = {}
                    temp[series] = dataset[series]
                    reqs.append(
                        ImportService._makeRequest(projectId, deviceId, temp))
                # Split series into chunks of up to _BATCH_SIZE
                else:
                    idx = 0
                    valsList = list(dataset[series])
                    while idx < seriesLen:
                        end = idx + ImportService._BATCH_SIZE
                        vals = valsList[idx:end]
                        temp = {}
                        temp[series] = vals
                        reqs.append(ImportService._makeRequest(
                            projectId, deviceId, temp))
                        idx += ImportService._BATCH_SIZE

        return reqs

    @staticmethod
    def _makeListOfBatchReqs(projectId, deviceId, dataBatch):
        """Creates a list of import requests from a DataStore.

        If the data set is under _BATCH_SIZE, it will be one request. Otherwise
        it will be split into multiple requests of at most _BACH_SIZE data
        points.

        Params:
            projectId - Project ID of the requests
            deviceId - Device ID of the requests
            dataset - The data batch to create requests from

        Returns:
            A list of batch import request bodies.
        """
        totalLen = len(dataBatch)

        reqs = []
        # No series, no requests
        if totalLen == 0:
            pass
        # Everything can fit in one request
        elif totalLen <= ImportService._BATCH_SIZE:
            reqs.append(ImportService._makeBatchRequest(projectId, deviceId, dataBatch))
        # Need to create multiple requests
        else:
            maxRowsPerBatch = ImportService._BATCH_SIZE // len(dataBatch.columns())
            batches = dataBatch.split(maxRowsPerBatch)
            for b in batches:
                reqs.append(ImportService._makeBatchRequest(projectId, deviceId, b))

        return reqs

    def importData(self, projectId, deviceId, dataSeries):
        """Wraps API call `POST /imports`

        Sends data to the iobeam backend to be stored.

        Params:
            projectId - Project ID the data belongs to
            deviceId - Device ID the data belongs to
            dataSeries - Dataset to send, as a dictionary where the keys are
            the name of the series, and the values are sets containing
            `iobeam.iobeam.DataPoint`s.

        Returns:
            A tuple where the first item is the success of all of the requests (True if
            all succeed, False otherwise); the second item is any error message or None
            if successful.

        Raises:
            Exception - If any of projectId, deviceId, or dataSeries is None.
        """
        if not self.token:
            raise request.UnauthorizedError.noTokenSet()
        if projectId is None:
            raise Exception("Project ID cannot be None")
        elif deviceId is None:
            raise Exception("Device ID cannot be None")
        elif dataSeries is None:
            raise Exception("Dataset cannot be None")
        elif len(dataSeries) == 0:
            return (True, None)
        endpoint = self.makeEndpoint("imports")

        reqs = ImportService._makeListOfReqs(projectId, deviceId, dataSeries)
        success = True
        extra = None
        for req in reqs:
            r = self.requester().post(endpoint).token(self.token)
            r.setBody(req)
            r.execute()
            success = success and (r.getResponseCode() == 200)
            if r.getResponseCode() != 200:
                extra = r.getResponse()

        return (success, extra)

    def importBatch(self, projectId, deviceId, dataStore):
        """Wraps API call `POST /imports?fmt=table`

        Sends data to the iobeam backend to be stored.

        Params:
            projectId - Project ID the data belongs to
            deviceId - Device ID the data belongs to
            dataStore - A `DataStore` object containing the the data to be imported

        Returns:
            A tuple where the first item is the success of all of the requests (True if
            all succeed, False otherwise); the second item is any error message or None
            if successful.

        Raises:
            ValueError - If validity checks fail for the token, project id, or device id.
        """
        utils.checkValidProjectId(projectId)
        utils.checkValidProjectToken(self.token)
        utils.checkValidDeviceId(deviceId)
        if dataStore is None or len(dataStore) == 0:
            utils.getLogger().warning("Attempted to send with no data")
            return (True, None)
        endpoint = self.makeEndpoint("imports")

        reqs = ImportService._makeListOfBatchReqs(projectId, deviceId, dataStore)
        success = True
        extra = None
        for req in reqs:
            r = self.requester().post(endpoint).token(self.token) \
                .setParam("fmt", "table") \
                .setBody(req)
            r.execute()
            success = success and (r.getResponseCode() == 200)
            if r.getResponseCode() != 200:
                extra = r.getResponse()

        return (success, extra)



"""Data types related to making data points and series."""
# pylint: disable=too-few-public-methods
from time import time
from enum import Enum
from iobeam.utils import utils

# For compatibility with both Python 2 and 3.
# pylint: disable=redefined-builtin,invalid-name
if utils.IS_PY3:
    long = int
    xrange = range
# pylint: enable=redefined-builtin,invalid-name


class TimeUnit(Enum):
    """Enum of different time units."""
    MILLISECONDS = "msec"
    MICROSECONDS = "usec"
    SECONDS = "sec"

class Timestamp(object):
    """Represents a timestamp, using a value and TimeUnit."""

    def __init__(self, value, unit=TimeUnit.MILLISECONDS):
        """Constructor of a Timestamp.

        Params:
            value - Numeric value of the timestamp. Must be an integer.
                    Precision is supported by using a different TimeUnit,
                    i.e., 5.5 seconds should be created as 5500 milliseconds
                    or 5500000 microseconds.
            unit - TimeUnit to use for the given value
        """
        if value is None or not isinstance(value, (int, long)):
            raise ValueError("timestamp value must be an int")
        self._value = value
        self._type = unit

    def __eq__(self, other):
        if other is None or not isinstance(other, Timestamp):
            return False

        return self.asMicroseconds() == other.asMicroseconds()

    def asMilliseconds(self):
        """This timestamp value represented in milliseconds.

        Returns:
            Timestamp coverted to an integral number of milliseconds.
            For timestamps with TimeUnit.MICROSECONDS, this is the value / 1000;
            for TimeUnit.MILLISECONDS this is just the value; and for
            TimeUnit.SECONDS, this is value * 1000.
        """
        if self._type == TimeUnit.MICROSECONDS:
            return int(self._value // 1000)
        elif self._type == TimeUnit.MILLISECONDS:
            return self._value
        elif self._type == TimeUnit.SECONDS:
            return (self._value * 1000)

    def asMicroseconds(self):
        """This timestamp value represented in microseconds.

        Returns:
            Timestamp converted to an integral number of microseconds.
            For timestamps with TimeUnit.MICROSECONDS, this is just the initial
            value; for TimeUnit.MILLISECONDS, value * 1000; and for
            TimeUnit.SECONDS, value * 1000000.
        """
        if self._type == TimeUnit.MICROSECONDS:
            return self._value
        elif self._type == TimeUnit.MILLISECONDS:
            return int(self._value * 1000)
        elif self._type == TimeUnit.SECONDS:
            return int(self._value * 1000000)
        else:
            raise ValueError("unknown type")


class DataPoint(object):
    """Represents a time-series datapoint, using a timestamp and value."""

    def __init__(self, value, timestamp=None):
        """Constructor of a DataPoint.

        Takes in a value, which must be a number type, and optionally a
        timestamp. If a timestamp is not supplied, the current time in
        milliseconds is used.

        Raises:
            ValueError - If value is not a number (int/float) or timestamp is not an
                         int.
        """
        if not isinstance(value, (int, long, float)):
            raise ValueError("'value' must be a number.")
        if timestamp is None:
            self._timestamp = Timestamp(int(time() * 1000),
                                        TimeUnit.MILLISECONDS)
        elif isinstance(timestamp, (int, long)):
            self._timestamp = Timestamp(timestamp, TimeUnit.MILLISECONDS)
        elif isinstance(timestamp, Timestamp):
            self._timestamp = timestamp
        else:
            raise ValueError(
                "invalid type for timestamp: {}".format(type(timestamp)))

        self._value = value

    def __str__(self):
        """Prints out: DataPoint{timestamp: <timestamp>, value: <value>}"""
        return "DataPoint{{timestamp: {}, value: {}}}".format(
            self._timestamp.asMicroseconds(), self._value)

    # pylint:disable=protected-access
    def __eq__(self, other):
        if other is None or not isinstance(other, DataPoint):
            return False
        return (self._value == other._value) and (
            self._timestamp == other._timestamp)
    # pylint:enable=protected-access

    def __hash__(self):
        return self._timestamp.asMicroseconds() + self._value

    def toDict(self):
        """Converts to a dictionary that is usable for the imports API.

        Returns:
            Dictionary with keys 'time' and 'value'.
        """
        return {
            "time": self._timestamp.asMicroseconds(),
            "value": self._value
        }


class DataStore(object):
    """A collection of data streams with rows batched by time."""

    def __init__(self, columns):
        """Construct a new DataStore object with given columns.

        Params:
            fields - Column or series names for data in this batch.

        Raises:
            ValueError - If `fields` is None, empty, or not a list.
        """
        if columns is None or len(columns) == 0:
            raise ValueError("columns cannot be None or empty")
        if not isinstance(columns, list):
            raise ValueError("columns must be a list of strings")
        self._columns = list(columns)  # defensive copy
        self._rows = []

    def clear(self):
        """Remove all data rows."""
        del self._rows[:]

    def add(self, timestamp, dataDict):
        """Add row of data at a given timestamp.

        Params:
            timestamp - Timestamp for all points to share
            dataDict - Data values keyed by which series they belong to.

        Raises:
            ValueError - For multiple cases:
                (a) timestamp is neither an int or a Timestamp type
                (b) dataDict is empty or None
                (b) dataDict contains keys not in this store
        """
        ts = None
        # validate timestamp
        if isinstance(timestamp, (int, long)):
            ts = Timestamp(timestamp)
        elif isinstance(timestamp, Timestamp):
            ts = timestamp
        else:
            raise ValueError("timestamp must be an int or Timestamp type")

        # validate data
        if dataDict is None or len(dataDict) == 0:
            raise ValueError("dataDict cannot be None or empty")
        for k in dataDict:
            if k not in self._columns:
                raise ValueError("dataDict can only contain keys in this store's columns")

        # everything ok, make row
        row = {"time": ts.asMicroseconds()}
        for f in self._columns:
            if f in dataDict:
                row[f] = dataDict[f]
            else:
                row[f] = None
        self._rows.append(row)

    def columns(self):
        """Return a copy of the columns in this store."""
        return list(self._columns)

    def rows(self):
        """Return a copy of the rows in this store."""
        ret = []
        for r in self._rows:
            ret.append(r.copy())
        return ret

    def split(self, chunkSize):
        """Split a store into multiple batches with `chunkSize` rows.

        Params:
            chunkSize - Max number of rows to include in a split

        Returns:
            List of DataStores containing at most chunkSize rows from this store.
        """
        ret = []
        for i in range(0, len(self._rows), chunkSize):
            temp = DataStore(self._columns)
            temp._rows = self._rows[i:i+chunkSize]
            ret.append(temp)

        return ret

    def __len__(self):
        """Return the size of this store in terms of data points."""
        return len(self._rows) * len(self._columns)


class DataSeries(object):
    """A collection of DataPoints for a given named series."""

    def __init__(self, name, points):
        """Constructor for DataSeries with a name and its points.

        Params:
            name - Name of the data series
            points - DataPoints in the series

        Throws:
            ValueError - If `name` is None
        """
        if name is None:
            raise ValueError("Name cannot be None")
        self._name = name
        self._points = list(points) if points is not None else []

    def __len__(self):
        """Returns the number of points in the series"""
        return len(self._points)

    def __str__(self):
        """Prints out: DataSeries{name: <name>, len: <size of points>}"""
        return "DataSeries{{name: {}, len: {}}}".format(self._name, len(self))

    def getName(self):
        """Return name of the series."""
        return self._name

    def getPoints(self):
        """Return the `DataPoint`s in the series."""
        return self._points

def makeUniformDataSeries(seriesName, startTime, endTime, values):
    """Creates a DataSeries from a list of values at uniform time intervals.

    Given a `startTime` and `endTime`, this function creates a DataSeries where
    each DataPoint is separated by a uniform time interval, with the first point
    at `startTime` and the last one at `endTime`. For example, if
    `startTime` is 0 and `endTime is 100, and there are 5 values, the resulting
    DataPoints would be at times 0, 25, 50, 75, 100.

    Params:
        seriesName - Name of the resulting DataSeries
        startTime - Time in milliseconds when the series should start
        endTime - Time in milliseconds when the series should end
        values - Numerical values to be made into DataPoints.

    Returns:
        A uniformly distributed DataSeries with DataPoints corresponding to
        `values`. None if the parameters are invalid, i.e., None or
        endTime <= startTime.
    """
    if (seriesName is None) or (startTime is None) or (endTime is None):
        return None
    elif not isinstance(startTime, (int, long)) or not isinstance(endTime, (int, long)):
        return None  # TODO: should raise ValueError?
    elif endTime <= startTime:
        return None  # TODO: should raise ValueError?
    elif (values is None) or len(values) == 0:
        return DataSeries(seriesName, [])

    valLen = len(values)
    if valLen == 1:
        pt = DataPoint(values[0], startTime)
        return DataSeries(seriesName, [pt])
    elif valLen == 2:
        pt1 = DataPoint(values[0], startTime)
        pt2 = DataPoint(values[1], endTime)
        return DataSeries(seriesName, [pt1, pt2])

    interval = (endTime - startTime) / (valLen - 1)
    pts = []

    for i in xrange(0, valLen - 1):
        timestamp = int(startTime + (i * interval))
        d = DataPoint(values[i], timestamp=timestamp)
        pts.append(d)
    pts.append(DataPoint(values[valLen - 1], timestamp=endTime))

    return DataSeries(seriesName, pts)



"""Used to communicate with iobeam's Tokens API"""
from iobeam.endpoints import service
from iobeam.http import request
from iobeam.utils import utils


class TokenService(service.EndpointService):
    """Communicates with the backend and exposes available Tokens API methods."""

    def __init__(self, requester=None):
        service.EndpointService.__init__(self, None, requester=requester)

    def refreshToken(self, oldToken):
        """Wraps API call `POST /tokens/project`

        Refreshes project token using an old token.

        Params:
            oldToken - Previous project token to refresh

        Returns:
            Refreshed project token (string), or None if an error occurs.
        """
        utils.checkValidProjectToken(oldToken)
        endpoint = self.makeEndpoint("tokens/project")
        req = {"refresh_token": oldToken}

        r = self.requester().post(endpoint).setBody(req)
        r.execute()

        if r.getResponseCode() == 200:
            return r.getResponse()["token"]
        else:
            return None



"""Common utility functions."""
import jwt
import logging
import sys
import time

IS_PY3 = sys.version_info >= (3, 0)

# For compatibility with both Python 2 and 3.
# pylint: disable=redefined-builtin,invalid-name
if IS_PY3:
    unicode = str
    long = int
# pylint: enable=redefined-builtin,invalid-name

EXPIRY_FUDGE = 1000 * 60 * 60 * 24  # one day in milliseconds

def isExpiredToken(projectToken):
    """Check if token's expiry date (minus a bit) has passed.

    Token is expired if the current time is after the expiry date less EXPIRY_FUDGE.

    Params:
        projectToken - JWT token to decode

    Returns:
        True if expired; False otherwise.

    Raises:
        ValueError - projectToken is not a valid JWT token
    """
    checkValidProjectToken(projectToken)
    opts = {"verify_signature": False, "verify_exp": False}
    try:
        decoded = jwt.decode(projectToken, '', options=opts)
    except jwt.DecodeError:
        raise ValueError("invalid jwt token")
    exp = int(decoded["exp"]) * 1000
    now = int(time.time() * 1000)

    return now >= (exp - EXPIRY_FUDGE)


def checkValidProjectId(projectId):
    """Check that a projectId is valid: a position int.

    Raises:
        ValueError - If projectId is None, not an int, or <= 0.
    """
    if projectId is None or not isinstance(projectId, (int, long)):
        raise ValueError("projectId must be an int")
    elif not projectId > 0:
        raise ValueError("projectId must be greater than 0")

def __checkNon0LengthString(value, valueName):
    """Check that a value is a non-0 length string.

    Raises:
        ValueError - If `token` is (a) None, (b) not a string/unicode, or
                     (c) is of length 0.
    """
    if value is None:
        raise ValueError("{} must be a string".format(valueName))
    elif not isinstance(value, (str, unicode)):
        raise ValueError("{} must be a string".format(valueName))
    elif len(value) == 0:
        raise ValueError("{} must be more than 0 characters".format(valueName))

def checkValidDeviceId(deviceId):
    """Check that a deviceId is valid: string of len > 0

    Raises:
        ValueError - If `deviceId` is (a) None, (b) not a string/unicode, or
                     (c) is of length 0.
    """
    __checkNon0LengthString(deviceId, "deviceId")

def checkValidProjectToken(token):
    """Check that a token is valid: string of len > 0

    Raises:
        ValueError - If `token` is (a) None, (b) not a string/unicode, or
                     (c) is of length 0.
    """
    __checkNon0LengthString(token, "token")

__LOGGER = None

def getLogger():
    global __LOGGER
    if __LOGGER is not None:
        logger = __LOGGER
    else:
        logger = logging.getLogger("iobeam")
        logger.setLevel(logging.WARNING)
        __LOGGER = logger

    if len(logger.handlers) == 0:
        logger.addHandler(logging.NullHandler())
    return logger



"""Used to communicate with iobeam's Devices API"""
from iobeam.endpoints import service
from iobeam.http import request
from iobeam.resources import device


class DuplicateIdError(request.Error):
    """Error for trying to register a device id that already exists."""

    def __init__(self):
        request.Error.__init__(self, "Device ID already registered.")


class DeviceService(service.EndpointService):
    """Communicates with the backend and exposes available Devices API methods."""

    def __init__(self, token, requester=None):
        service.EndpointService.__init__(self, token, requester=requester)

    def getTimestamp(self):
        """Wraps API call `GET /devices/timestamp`.

        This returns the current time in milliseconds, according to the iobeam
        backend. Useful for clients with limited on-device clock support.

        Returns:
            Current timestamp in milliseconds since epoch; -1 if error.
        """
        if not self.token:
            raise request.UnauthorizedError.noTokenSet()
        endpoint = self.makeEndpoint("devices/timestamp")

        r = self.requester().get(endpoint).token(self.token)
        r.execute()

        if r.getResponseCode() == 200:
            resp = r.getResponse()
            return resp["server_timestamp"]
        else:
            return -1

    def registerDevice(self, projectId, deviceId=None, deviceName=None):
        """Wraps API call `POST /devices`

        Registers the device in project `projectId` with the iobeam backend.
        `deviceId` and `deviceName` are optional, if provided they will used
        in registration, otherwise they will be generated by the backend.

        Params:
            projectId - Project ID to register device in
            deviceId - Desired device ID; if None, will be generated.
            deviceName - Desired device name; if None, will be generated.

        Returns:
            A Device object corresponding to the parameters (explicit and
            generated); None if there is an error/failure.
        """
        if not self.token:
            raise request.UnauthorizedError.noTokenSet()
        endpoint = self.makeEndpoint("devices")

        r = self.requester().post(endpoint).token(self.token)
        reqBody = {"project_id": projectId}
        if deviceId or deviceName:
            if deviceId:
                reqBody["device_id"] = deviceId
            if deviceName:
                reqBody["device_name"] = deviceName
        r.setBody(reqBody)
        r.execute()

        ret = None
        if r.getResponseCode() == 201:
            resp = r.getResponse()
            ret = device.Device(projectId, resp["device_id"],
                                deviceName=resp["device_name"])
        elif r.getResponseCode() == 403:
            raise request.UnauthorizedError("Invalid credentials.")
        elif r.getApiErrorCode() == request.ERROR_CODE_DUPLICATE_DEVICE_ID:
            raise DuplicateIdError()
        else:
            raise request.Error(
                    "Received unexpected code: {}".format(r.getResponseCode()))

        return ret



"""Data types related to iobeam devices."""
from iobeam.utils import utils


# pylint: disable=too-few-public-methods
class Device(object):
    """Represents a device that is associated with a project."""

    def __init__(self, projectId, deviceId, deviceName=None):
        """Constructor for a Device object.

        A valid Device object has at least a `projectId` and a `deviceId`, and
        optionally a `deviceName`.

        Params:
            projectId - Project id (int) that this device belongs to
            deviceId - Id of this device
            deviceName - Optional secondary identifier for the device

        Raises:
            ValueError - If projectId is None, not an int, or not >0. Also if
                         deviceId is None.
        """
        utils.checkValidProjectId(projectId)
        utils.checkValidDeviceId(deviceId)

        self.projectId = projectId
        self.deviceId = deviceId
        self.deviceName = deviceName

    def __str__(self):
        """Prints a string representation of this device.

        Format:
        <deviceName> [<projectId>: <deviceId>]
        """
        return "{} [{}: {}]".format(self.deviceName, self.projectId,
                                    self.deviceId)
# pylint: enable=too-few-public-methods



"""Classes used when communicating via HTTP to the iobeam backend."""
import requests

ERROR_CODE_DUPLICATE_DEVICE_ID = 150

class Requester(object):
    """Generates HTTP requests using `requests` library."""

    _BASE_URL = "https://api.iobeam.com/v1/"

    def __init__(self, baseUrl=_BASE_URL):
        self._baseUrl = baseUrl
        self._session = requests.Session()

    def makeEndpoint(self, endpoint):
        """Create a fully defined URL for an endpoint."""
        return self._baseUrl + endpoint

    def get(self, url):
        """Return a base GET request for a given URL."""
        return Request("GET", url, self._session)

    def post(self, url):
        """Return a base POST request for a given URL."""
        return Request("POST", url, self._session)


class UnauthorizedError(Exception):
    """Error for forbidden access."""

    def __init__(self, value):
        Exception.__init__(self, value)

    @staticmethod
    def noTokenSet():
        """Return an `UnauthorizedError` when a token is not set."""
        return UnauthorizedError("no token set.")

class Error(Exception):
    """Generic error for HTTP operations."""

    def __init__(self, value):
        Exception.__init__(self, value)

class Request(object):
    """Wrapper for an HTTP request object."""

    def __init__(self, method, url, session):
        self.method = method
        self.url = url
        self.headers = {}
        self.header("User-Agent", "iobeam python")
        self.header("Connection", "keep-alive")
        self.resp = None
        self.body = None
        self.params = {}
        self._session = session

    def header(self, key, value):
        """Add a header to the request (chainable)."""
        self.headers[key] = value
        return self

    def token(self, token):
        """Set a token for the request (chainable)."""
        return self.header("Authorization", "Bearer {}".format(token))

    def setBody(self, body):
        """Set the request body (chainable)."""
        self.body = body
        return self

    def setParam(self, key, value):
        """Set a query parameter for a request (chainable)."""
        self.params[key] = value
        return self

    def execute(self):
        """Execute an HTTP request using `requests` library."""
        self.resp = None
        if self.method == "GET":
            self.resp = self._session.get(
                self.url, params=self.params, headers=self.headers)
        elif self.method == "POST":
            self.resp = self._session.post(self.url, params=self.params,
                                           headers=self.headers, json=self.body)
        else:
            print("unsupported method")

    def getResponse(self):
        """Return response body for a given request."""
        if self.resp is None:
            return None
        try:
            return self.resp.json()
        except Exception:
            return None

    def getResponseCode(self):
        """Return HTTP status code for a given request."""
        return self.resp.status_code

    def getApiErrorCode(self):
        """Return iobeam API error code.

        Returns:
            The error code or None otherwise.
        """
        if self.resp is None:
            return None
        try:
            msg = self.resp.json()
            return msg["errors"][0]["code"]
        except Exception:
            return None



"""Iobeam client library."""
from . import iobeam

ClientBuilder = iobeam.ClientBuilder
DataStore = iobeam.DataStore
DataPoint = iobeam.DataPoint
DataSeries = iobeam.DataSeries
Timestamp = iobeam.Timestamp
TimeUnit = iobeam.TimeUnit
QueryReq = iobeam.QueryReq

# pylint:disable=invalid-name
makeQuery = iobeam.makeQuery
# pylint:enable=invalid-name



"""The iobeam client and related types/methods."""
from .endpoints import devices
from .endpoints import exports
from .endpoints import imports
from .endpoints import tokens
from .http import request
from .resources import data
from .resources import device
from .resources import query
from .utils import utils

import os.path

#  Aliases for resource types for convenience outside the package.
DataStore = data.DataStore
DataPoint = data.DataPoint
DataSeries = data.DataSeries
Timestamp = data.Timestamp
TimeUnit = data.TimeUnit
QueryReq = query.Query

_DEVICE_ID_FILE = "iobeam_device_id"

class ClientBuilder(object):
    """Used to build an iobeam client object."""

    def __init__(self, projectId, projectToken):
        utils.checkValidProjectId(projectId)
        utils.checkValidProjectToken(projectToken)

        self._projectId = projectId
        self._projectToken = projectToken
        self._diskPath = None
        self._deviceId = None
        self._regArgs = None
        self._backend = None

    def saveToDisk(self, path="."):
        """Client object should save deviceId to disk (chainble).

        Params:
            path - File system path to save deviceId

        Returns:
            This Builder object, for chaining.
        """
        self._diskPath = path
        return self

    def setDeviceId(self, deviceId):
        """Client object should set deviceId (chainable)."""
        utils.checkValidDeviceId(deviceId)
        self._deviceId = deviceId
        return self

    def registerOrSetId(self, deviceId):
        """Client object should register itself, or set the id if it exists.

        Params:
            deviceId - Desired device id to register or set if it exists

        Returns:
            This Builder object, for chaining.
        """
        utils.checkValidDeviceId(deviceId)
        self._regArgs = (deviceId, None, True)
        return self


    def registerDevice(self, deviceId=None, deviceName=None):
        """Client object should register itself (chainable).

        Params:
            deviceId - Desired device id (optional)
            deviceName - Desired deviceName (optional)

        Returns:
            This Builder object, for chaining.
        """
        if deviceId is not None:
            utils.checkValidDeviceId(deviceId)
        self._regArgs = (deviceId, deviceName, False)
        return self

    def setBackend(self, baseUrl):
        """Client object should use this url as the backend (chainable).

        Params:
            baseUrl - Base part of the URL to use for making API requests.

        Returns:
            This Builder object, for chaining.
        """
        self._backend = request.Requester(baseUrl=baseUrl)
        return self

    def build(self):
        """Actually construct the client object."""
        client = _Client(self._diskPath, self._projectId, self._projectToken,
                         self._backend, deviceId=self._deviceId)
        if self._regArgs is not None:
            did, dname, setOnDupe = self._regArgs
            client.registerDevice(deviceId=did, deviceName=dname,
                                  setOnDupe=setOnDupe)

        return client


class _Client(object):
    """Client object used to communicate with iobeam."""

    # pylint: disable=too-many-arguments
    def __init__(self, path, projectId, projectToken, backend, deviceId=None):
        """Constructor for iobeam client object.

        Creates a client instance associated with a project and (potentially) a
        device. If `path` is provided, this device's ID will be stored at
        <path>/iobeam_device_id. This on-disk ID will be used if one is not
        provided as `deviceId`.

        Params:
            path - Path where device ID should be persisted
            projectId - iobeam project ID
            projectToken - iobeam project token with write access for sending data
            backend - Base url of the backend to use; if None, requests go to
                      https://api.iobeam.com/v1/
            deviceId - Device id if previously registered
        """
        utils.checkValidProjectId(projectId)
        utils.checkValidProjectToken(projectToken)

        self.projectId = projectId
        self.projectToken = projectToken
        self._path = path
        self._dataset = {}
        self._batches = []

        self._activeDevice = None
        if deviceId is not None:
            self._setActiveDevice(device.Device(projectId, deviceId))
        elif self._path is not None:
            p = os.path.join(self._path, _DEVICE_ID_FILE)
            if os.path.isfile(p):
                with open(p, "r") as f:
                    did = f.read()
                    if len(did) > 0:
                        self._activeDevice = device.Device(projectId, did)


        # Setup services
        self._deviceService = devices.DeviceService(projectToken,
                                                    requester=backend)
        self._importService = imports.ImportService(projectToken,
                                                    requester=backend)
        self._tokenService = tokens.TokenService(requester=backend)

        self._checkToken()
    # pylint: enable=too-many-arguments

    def _checkToken(self):
        if utils.isExpiredToken(self.projectToken):
            newToken = self._refreshToken()
            if newToken is not None:
                self.projectToken = newToken


    def _refreshToken(self):
        return self._tokenService.refreshToken(self.projectToken)

    def registerDevice(self, deviceId=None, deviceName=None, setOnDupe=False):
        """Registers the device with iobeam.

        If a path was provided when the client was constructed, the device ID
        will be stored on disk.

        Params:
            deviceId - Desired device ID; otherwise randomly generated
            deviceName - Desired device name; otherwise randomly generated
            setOnDupe - If duplicate device id, use the id instead of raising an
                        error; default False (will throw an error if duplicate).

        Returns:
            This client object (allows for chaining)

        Raises:
            devices.DuplicateIdError - If id is a dupliecate and `setOnDupe` is
                                       False.
        """
        activeSet = self._activeDevice is not None
        didIsNone = deviceId is None
        if activeSet and (didIsNone or self._activeDevice.deviceId == deviceId):
            return self

        if deviceId is not None:
            utils.checkValidDeviceId(deviceId)

        self._checkToken()
        try:
            d = self._deviceService.registerDevice(self.projectId,
                                                   deviceId=deviceId,
                                                   deviceName=deviceName)
        except devices.DuplicateIdError:
            if setOnDupe:
                d = device.Device(self.projectId, deviceId,
                                  deviceName=deviceName)
            else:
                raise
        self._setActiveDevice(d)

        return self

    def setDeviceId(self, deviceId):
        """Set client's active device id."""
        d = device.Device(self.projectId, deviceId)
        self._setActiveDevice(d)

    def getDeviceId(self):
        """Get client's active device id."""
        if self._activeDevice is None:
            return None
        else:
            return self._activeDevice.deviceId

    def _setActiveDevice(self, dev):
        """Internally sets the client's device id, including saving to disk."""
        self._activeDevice = dev
        if self._path is not None:
            p = os.path.join(self._path, _DEVICE_ID_FILE)
            with open(p, "w") as f:
                f.write(self._activeDevice.deviceId)

    def isRegistered(self):
        """Tells whether this client has a registered device.

        Returns:
            True if there is a device ID associated with this client
        """
        return self._activeDevice is not None

    def addDataPoint(self, seriesName, datapoint):
        """Adds a single DataPoint to a series in the data store.

        If the series does not currently exist, it will be created.

        Params:
            seriesName - The series to add the datapoint to
            datapoint - DataPoint to be added

        Raises:
            ValueError - If series name is None, not a string, or length 0.
        """
        if seriesName is None or not isinstance(seriesName, str):
            raise ValueError("seriesName cannot be None or a non-string")
        elif len(seriesName) == 0:
            raise ValueError("seriesName cannot be a 0-length string")
        elif datapoint is None or not isinstance(datapoint, data.DataPoint):
            utils.getLogger().warning("tried to add an invalid or None datapoint")
            return

        if seriesName not in self._dataset:
            self._dataset[seriesName] = set()
        self._dataset[seriesName].add(datapoint)

    def addDataSeries(self, dataseries):
        """Adds a DataSeries to the data store.

        If the series exists, all the points will be added.

        Params:
            dataseries - The DataSeries to add to the data store.

        Raises:
            ValueError - If dataseries is None or not a DataSeries object
        """
        if dataseries is None or not isinstance(dataseries, data.DataSeries):
            raise ValueError("dataseries was None or not a DataSeries")
        elif len(dataseries) == 0:
            utils.getLogger().warning("tried to add empty dataseries")
            return

        key = dataseries.getName()
        if key not in self._dataset:
            self._dataset[key] = set()

        for p in dataseries.getPoints():
            self._dataset[key].add(p)

    def clearSeries(self, seriesName):
        """Removes any points associated with `seriesName`."""
        self._dataset.pop(seriesName, None)

    def createDataStore(self, columns):
        """Create a DataStore that is tracked by this client.

        Params:
            columns - List of stream names for the DataStore

        Returns:
            DataStore object with those columns and being tracked
            by this client for sending.
        """
        ds = data.DataStore(columns)
        self._batches.append(ds)

        return ds

    def addDataStore(self, store):
        """Add a DataStore to this client.

        Params:
            store - The DataStore to add to the data store.
        """
        if store is None:
            utils.getLogger().warning("adding store was None")
            return
        elif not isinstance(batch, data.DataStore):
            raise ValueError("store must be a DataStore")
        elif len(store) == 0:
            return

        self._batches.append(store)

    def _convertDataSetToBatches(self):
        dataset = self._dataset
        batches = []
        for name in dataset:
            batch = data.DataStore([name])
            for point in dataset[name]:
                asDict = point.toDict()
                ts = data.Timestamp(asDict["time"], unit=TimeUnit.MICROSECONDS)
                row = {}
                row[name] = asDict["value"]
                batch.add(ts, row)
            batches.append(batch)
        self._dataset = {}
        return batches

    def send(self):
        """Sends stored data to the iobeam backend.

        Raises:
            Exception - if sending the data fails.
        """
        self._checkToken()
        pid = self.projectId
        did = self._activeDevice.deviceId
        tempBatches = self._convertDataSetToBatches()

        for b in list(self._batches):
            success, extra = self._importService.importBatch(pid, did, b)
            if not success:
                raise Exception("send failed. server sent: {}".format(extra))
            else:
                b.clear()

        # temp batches are not saved between calls; re-made each time
        for b in tempBatches:
            success, extra = self._importService.importBatch(pid, did, b)
            if not success:
                raise Exception("send failed. server sent: {}".format(extra))


    @staticmethod
    def query(token, qry, backend=None):
        """Performs a query on the iobeam backend.

        The Query specifies the project, device, and series to look up, as well
        as any parameters to use.

        Params:
            token - A token with read access for the given project.
            query - Specifies a data query to perform.

        Returns:
            A dictionary representing the results of the query.

        Raises:
            ValueError - If `token` or `query` is None, or `query` is the wrong
                         type.
        """
        if token is None:
            raise ValueError("token cannot be None")
        elif qry is None:
            raise ValueError("qry cannot be None")
        elif not isinstance(qry, QueryReq):
            raise ValueError("qry must be a iobeam.QueryReq")
        requester = None
        if backend is not None:
            requester = request.Requester(baseUrl=backend)

        service = exports.ExportService(token, requester=requester)
        return service.getData(qry)


# Alias
def makeQuery(token, qry, backend=None):
    """Perform iobeam query."""
    return _Client.query(token, qry, backend)
